"""Tool to build optimized buckets out of Quantitative and Qualitative features
for continuous regression tasks.
"""

from typing import Callable

from numpy import mean, unique
from pandas import DataFrame, Series
from scipy.stats import kruskal

from ..discretizers import GroupedList
from ..discretizers.utils.base_discretizers import extend_docstring
from .base_carver import BaseCarver


class ContinuousCarver(BaseCarver):
    """Automatic carving of continuous, discrete, categorical and ordinal
    features that maximizes association with a continuous target.

    For continuous targets, :ref:`Kruskal` is used as association measure to sort combinations.
    """

    @extend_docstring(BaseCarver.__init__)
    def __init__(
        self,
        min_freq: float,
        *,
        quantitative_features: list[str] = None,
        qualitative_features: list[str] = None,
        ordinal_features: list[str] = None,
        values_orders: dict[str, GroupedList] = None,
        max_n_mod: int = 5,
        min_freq_mod: float = None,
        output_dtype: str = "float",
        dropna: bool = True,
        copy: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        """
        # association measure used to find the best groups for continuous targets
        if "sort_by" in kwargs:
            assert kwargs.get("sort_by") == "kruskal", (
                f" - [ContinuousCarver] Measure '{kwargs.get('sort_by')}' not yet implemented for "
                "continuous targets. Only 'kruskal' measure is implemented for continuous targets."
            )
            kwargs.pop("sort_by")

        # Initiating BaseCarver
        super().__init__(
            min_freq=min_freq,
            sort_by="kruskal",
            quantitative_features=quantitative_features,
            qualitative_features=qualitative_features,
            ordinal_features=ordinal_features,
            values_orders=values_orders,
            max_n_mod=max_n_mod,
            min_freq_mod=min_freq_mod,
            output_dtype=output_dtype,
            dropna=dropna,
            copy=copy,
            verbose=verbose,
            **kwargs,
        )

    def _prepare_data(
        self,
        X: DataFrame,
        y: Series,
        X_dev: DataFrame = None,
        y_dev: Series = None,
    ) -> tuple[DataFrame, DataFrame]:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in ``AutoCarver.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        X_dev : DataFrame, optional
            Dataset to evalute the robustness of discretization, by default ``None``
            It should have the same distribution as X.

        y_dev : Series, optional
            Binary target feature with wich the robustness of discretization is evaluated, by default ``None``

        Returns
        -------
        tuple[DataFrame, DataFrame]
            Copies of (X, X_dev) to be used according to target type
        """
        # Checking for binary target and copying X
        x_copy, x_dev_copy = super()._prepare_data(X, y, X_dev=X_dev, y_dev=y_dev)

        # continuous target, checking values
        y_values = unique(y)
        assert (
            len(y_values) > 2
        ), " - [ContinuousCarver] provided y is binary, consider using BinaryCarver instead."
        not_numeric = str in y.apply(type).unique()
        assert (
            not not_numeric
        ), " - [ContinuousCarver] y must be a continuous Series (int or float, not object)"

        return x_copy, x_dev_copy

    def _aggregator(
        self, features: list[str], X: DataFrame, y: Series, labels_orders: dict[str, GroupedList]
    ) -> dict[str, DataFrame]:
        """Computes y values for modalities of specified features and ensures the ordering according to the known labels

        Parameters
        ----------
        features : list[str]
            _description_
        X : DataFrame
            _description_
        y : Series
            _description_
        labels_orders : dict[str, GroupedList]
            _description_

        Returns
        -------
        dict[str, DataFrame]
            _description_
        """
        # checking for empty datasets
        yvals = {feature: None for feature in features}
        if X is not None:
            # crosstab for each feature
            for feature in features:
                # computing crosstab with str_nan
                yval = y.groupby(X[feature]).apply(lambda u: list(u))

                # reordering according to known_order
                yval = yval.reindex(labels_orders[feature])

                # storing results
                yvals.update({feature: yval})

        return yvals

    def _grouper(self, yval: Series, groupby: dict[str:str]) -> Series:
        """Groups values of y

        Parameters
        ----------
        yval : Series
            _description_
        groupby : _type_
            _description_

        Returns
        -------
        Series
            _description_
        """
        # TODO: try to convert this to the vectorial version like BinaryCarver
        return yval.groupby(groupby).sum()

    def _association_measure(self, yval: Series, **kwargs) -> dict[str, float]:
        """Computes measures of association between feature and quantitative target.

        Parameters
        ----------
        yval : DataFrame
            Values taken by y for each of x's modalities.

        Returns
        -------
        dict[str, float]
            Kruskal-Wallis' H as a dict.
        """
        # Kruskal-Wallis' H
        return {"kruskal": kruskal(*tuple(yval.values))[0]}

    def _target_rate(self, yval: Series) -> Series:
        """Computes target average per group for a continuous target

        Parameters
        ----------
        yval : Series
            _description_

        Returns
        -------
        Series
            _description_
        """
        return yval.apply(mean).sort_values()

    def _printer(self, yval: Series = None) -> DataFrame:
        """Prints a continuous yval's statistics

        Parameters
        ----------
        yval : Series
            A series of values of y by modalities of x, by default None

        Returns
        -------
        DataFrame
            Target rate and frequency per modality
        """
        # checking for an xtab
        stats = None
        if yval is not None:
            # target rate and frequency statistics per modality
            stats = DataFrame(
                {
                    # target rate per modality
                    "target_rate": yval.apply(mean),
                    # frequency per modality
                    "frequency": yval.apply(len) / yval.apply(len).sum(),
                }
            )

        return stats

    @extend_docstring(BaseCarver.fit)
    def fit(
        self,
        X: DataFrame,
        y: Series,
        *,
        X_dev: DataFrame = None,
        y_dev: Series = None,
    ) -> None:
        # preparing datasets and checking for wrong values
        x_copy, x_dev_copy = self._prepare_data(X, y, X_dev, y_dev)

        # Fitting BaseCarver
        super().fit(x_copy, y, X_dev=x_dev_copy, y_dev=y_dev)

        return self



from codecs import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    author_email="defrancemario@gmail.com",
    description="Automatic Discretization of Features with Optimal Target Association",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mdefrance/AutoCarver",
    project_urls={
        "Documentation": "https://autocarver.readthedocs.io/en/latest/index.html",
        "Bug Tracker": "https://github.com/mdefrance/AutoCarver/issues",
    },
    license="MIT",
    install_requires=[
        "pandas",
        "numpy",
        "scipy",
        "scikit-learn",
        "statsmodels",
        "tqdm",
    ],
    extras_require={
        "jupyter": [
            "ipython",
        ]
    },
    packages=find_packages(),
    classifiers=[
        # ou 4 - Beta ou 5 - Production/Stable
        # "Development Status :: 5 - Production/Stable",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.9",
)



"""Tool to build optimized buckets out of Quantitative and Qualitative features
for multiclass classification tasks.
"""

from typing import Any, Callable

from pandas import DataFrame, Series, unique

from ..discretizers import BaseDiscretizer, GroupedList
from ..discretizers.utils.base_discretizers import extend_docstring
from .base_carver import BaseCarver
from .binary_carver import BinaryCarver


class MulticlassCarver(BaseCarver):
    """Automatic carving of continuous, discrete, categorical and ordinal
    features that maximizes association with a multiclass target.
    """

    @extend_docstring(BinaryCarver.__init__)
    def __init__(
        self,
        sort_by: str,
        min_freq: float,
        *,
        quantitative_features: list[str] = None,
        qualitative_features: list[str] = None,
        ordinal_features: list[str] = None,
        values_orders: dict[str, GroupedList] = None,
        max_n_mod: int = 5,
        min_freq_mod: float = None,
        output_dtype: str = "float",
        dropna: bool = True,
        copy: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        # association measure used to find the best groups for multiclass targets
        implemented_measures = ["tschuprowt", "cramerv"]
        assert sort_by in implemented_measures, (
            f" - [MulticlassCarver] Measure '{sort_by}' not yet implemented for multiclass targets"
            f". Choose from: {str(implemented_measures)}."
        )

        # Initiating BinaryCarver
        super().__init__(
            min_freq=min_freq,
            sort_by=sort_by,
            quantitative_features=quantitative_features,
            qualitative_features=qualitative_features,
            ordinal_features=ordinal_features,
            values_orders=values_orders,
            max_n_mod=max_n_mod,
            min_freq_mod=min_freq_mod,
            output_dtype=output_dtype,
            dropna=dropna,
            copy=copy,
            verbose=verbose,
            **kwargs,
        )
        self.kwargs = kwargs

    def _prepare_data(
        self,
        X: DataFrame,
        y: Series,
        X_dev: DataFrame = None,
        y_dev: Series = None,
    ) -> tuple[DataFrame, DataFrame, dict[str, Callable]]:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in ``AutoCarver.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        X_dev : DataFrame, optional
            Dataset to evalute the robustness of discretization, by default ``None``
            It should have the same distribution as X.

        y_dev : Series, optional
            Binary target feature with wich the robustness of discretization is evaluated, by default ``None``

        Returns
        -------
        tuple[DataFrame, DataFrame, dict[str, Callable]]
            Copies of (X, X_dev) and helpers to be used according to target type
        """
        # Checking for binary target and copying X
        x_copy, x_dev_copy = super()._prepare_data(X, y, X_dev=X_dev, y_dev=y_dev)

        # converting target as str
        y_copy = y.astype(str)

        # multiclass target, checking values
        y_values = unique(y_copy)
        assert (
            len(y_values) > 2
        ), " - [MulticlassCarver] provided y is binary, consider using BinaryCarver instead."

        y_dev_copy = y_dev
        if y_dev is not None:
            # converting target as str
            y_dev_copy = y_dev.astype(str)

            # check that classes of y are classes of y_dev
            unique_y_dev = y_dev.unique()
            unique_y = y.unique()
            assert all(mod_y in unique_y_dev for mod_y in unique_y), (
                "- [MulticlassCarver] Some classes of y are missing from y_dev: "
                f"{str([mod_y for mod_y in unique_y if mod_y not in unique_y_dev])}"
            )
            assert all(mod_y_dev in unique_y for mod_y_dev in unique_y_dev), (
                "- [MulticlassCarver] Some classes of y_dev are missing from y: "
                f"{str([mod_y_dev for mod_y_dev in unique_y_dev if mod_y_dev not in unique_y])}"
            )

        return x_copy, y_copy, x_dev_copy, y_dev_copy

    @extend_docstring(BinaryCarver.fit)
    def fit(
        self,
        X: DataFrame,
        y: Series,
        *,
        X_dev: DataFrame = None,
        y_dev: Series = None,
    ) -> None:
        # preparing datasets and checking for wrong values
        x_copy, y_copy, x_dev_copy, y_dev_copy = self._prepare_data(X, y, X_dev, y_dev)

        # getting distinct y classes
        y_classes = sorted(list(y_copy.unique()))[1:]  # removing one of the classes

        # copying raw values_orders (contains previous discretizations)
        raw_values_orders = {feature: order for feature, order in self.values_orders.items()}

        # initiating casted features input_dtypes and values_orders
        casted_values_orders: dict[str, GroupedList] = {}
        casted_input_dtypes: dict[str, str] = {}
        casted_features = {feature: [] for feature in self.features}

        # iterating over each class minus one
        for n, y_class in enumerate(y_classes):
            if self.verbose:  # verbose if requested
                print(
                    f"\n---------\n[MulticlassCarver] Fit y={y_class} ({n+1}/{len(y_classes)})\n------"
                )

            # identifying this y_class
            target_class = (y_copy == y_class).astype(int)
            target_class_dev = None
            if y_dev_copy is not None:
                target_class_dev = (y_dev_copy == y_class).astype(int)

            # initiating BinaryCarver for y_class
            binary_carver = BinaryCarver(
                min_freq=self.min_freq,
                sort_by=self.sort_by,
                quantitative_features=self.quantitative_features,
                qualitative_features=self.qualitative_features,
                ordinal_features=self.ordinal_features,
                values_orders=raw_values_orders,
                max_n_mod=self.max_n_mod,
                output_dtype=self.output_dtype,
                dropna=self.dropna,
                copy=True,  # copying x to keep raw columns as is
                verbose=self.verbose,
                **self.kwargs,
            )

            # fitting BinaryCarver for y_class
            binary_carver.fit(x_copy, target_class, X_dev=x_dev_copy, y_dev=target_class_dev)

            # renaming BinaryCarver's fitted values_orders/input_dtype/output_dtype for y_class
            casted_values_orders.update(dict_append_class(binary_carver.values_orders, y_class))
            casted_input_dtypes.update(dict_append_class(binary_carver.input_dtypes, y_class))
            self._history.update(dict_append_class(binary_carver._history, y_class))
            for feature in binary_carver.features:
                # feature only present in binary_carver.feature if not removed
                casted_features.update(
                    {feature: casted_features.get(feature) + [append_class(feature, y_class)]}
                )

            if self.verbose:  # verbose if requested
                print("---------\n")

        # initiating BaseDiscretizer with features_casting
        BaseDiscretizer.__init__(
            self,
            features=[feature for castings in casted_features.values() for feature in castings],
            values_orders=casted_values_orders,
            input_dtypes=casted_input_dtypes,
            output_dtype=self.output_dtype,
            str_nan=self.kwargs.get("str_nan", "__NAN__"),
            str_default=self.kwargs.get("str_default", "__OTHER__"),
            dropna=self.dropna,
            copy=self.copy,
            verbose=self.verbose,
            features_casting=casted_features,
        )

        # fitting BaseDiscretizer
        BaseDiscretizer.fit(self, x_copy, y_copy)

        return self


def append_class(string: str, to_append: str):
    """add the to_append string to a string"""

    return f"{string}_{to_append}"


def dict_append_class(dic: dict[str, Any], to_append: str):
    """add the to_append string to all keys if dic"""

    return {append_class(feature, to_append): values for feature, values in dic.items()}
